import * as path from "https://deno.land/std@0.213.0/path/mod.ts";
import { z } from "https://deno.land/x/zod@v3.22.4/mod.ts";

/**
 * Gets the root path of the clonned project.
 */
export function root_path() {
	const this_path = new URL(import.meta.url).pathname;
	return path.resolve(this_path, "..", "..");
}

/**
 * Creates handles to a channel.\
 * c.f; https://en.wikipedia.org/wiki/Channel_(programming)
 */
export function channel<T>() {
	const inner = {
		items: [] as T[],
		awaiters: [] as ((item: T) => void)[],
	};

	function send(item: T) {
		const [awaiter] = inner.awaiters.splice(0, 1);
		if (awaiter !== undefined) awaiter(item);
		else inner.items.push(item);
	}

	function receive() {
		const [item] = inner.items.splice(0, 1);
		if (item !== undefined) return resolves_to(item);
		else return new Promise<T>((resolver) => inner.awaiters.push(resolver));
	}

	return { send, receive };
}

const resolves_to = <T>(item: T) => new Promise<T>((r) => r(item));

/**
 * Returns both a promise and its resolver.
 */
export function split_promise<T>() {
	let resolver: null | ((item: T) => void) = null;
	const promise = new Promise<T>((r) => resolver = r);
	resolver = resolver as unknown as (item: T) => void;
	return { promise, resolver };
}

export type Resolver<T> = (item: T) => void;

/**
 * Logging function factory.
 */
export function log_from(url: string, ...args: unknown[]) {
	const date = new Date().toLocaleString("fr-FR");
	const file = path.basename(new URL(url).pathname);
	console.log(`[${date}][epitls][${file}]`, ...args);
}

export type Conf = Awaited<ReturnType<typeof read_conf>>;
/**
 * Reads and parse a configuration file containing the current instance' secrets.
 */
export async function read_conf(path: string) {
	const content = await Deno.readTextFile(path);
	return z.object({
		discord: z.object({
			bot_token: z.string(),
		}),
		cri: z.object({
			cri_token: z.string(),
		}),
		email_smtp: z.object({
			hostname: z.string(),
			port: z.number(),
			username: z.string(),
			password: z.string(),
		}),
		verif_http: z.object({
			port: z.number(),
			url_prefix: z.string(),
		}),
	}).parse(JSON.parse(content));
}

export function placeholder_conf() {
	const result: Conf = {
		"discord": {
			"bot_token": "available at 'https://discord.com/developers/applications'",
		},
		"cri": {
			"cri_token": "'username:password' base64-encoded",
		},
		"email_smtp": {
			"hostname": "smtp.gmail.com",
			"port": 465,
			"username": "bot@gmail.com",
			"password": "PWD",
		},
		"verif_http": {
			"port": 80,
			"url_prefix": "localhost",
		},
	};
	return JSON.stringify(result, null, 4);
}

export type SimpleResult = true | string;

export async function wait(ms: number) {
	const { promise, resolver } = split_promise<void>();
	setTimeout(resolver, ms);
	await promise;
}
